/*                                                                             
 * Copyright 2005,2006 WSO2, Inc. http://www.wso2.org
 *                                                                             
 * Licensed under the Apache License, Version 2.0 (the "License");             
 * you may not use this file except in compliance with the License.            
 * You may obtain a copy of the License at                                     
 *                                                                             
 *      http://www.apache.org/licenses/LICENSE-2.0                             
 *                                                                             
 * Unless required by applicable law or agreed to in writing, software         
 * distributed under the License is distributed on an "AS IS" BASIS,           
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    
 * See the License for the specific language governing permissions and         
 * limitations under the License.                                              
 */
package org.wso2.solutions.identity;

import java.util.ArrayList;
import java.util.Iterator;

import org.apache.axis2.AxisFault;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.solutions.identity.admin.ClaimsAdmin;
import org.wso2.solutions.identity.admin.ConfigurationContextHolder;
import org.wso2.solutions.identity.admin.ParameterAdmin;
import org.wso2.solutions.identity.admin.STSConfigAdmin;
import org.wso2.solutions.identity.i18n.Messages;
import org.wso2.solutions.identity.persistence.IPPersistenceManager;
import org.wso2.solutions.identity.persistence.dataobject.ActionDO;
import org.wso2.solutions.identity.persistence.dataobject.ClaimDO;
import org.wso2.solutions.identity.persistence.dataobject.DialectDO;
import org.wso2.solutions.identity.persistence.dataobject.RealmConfigurationDO;
import org.wso2.solutions.identity.persistence.dataobject.RealmConfigurationPropertyDO;
import org.wso2.solutions.identity.persistence.dataobject.RealmDO;
import org.wso2.solutions.identity.persistence.dataobject.RelyingPartyDO;
import org.wso2.solutions.identity.users.IdentityDefaultRealm;
import org.wso2.solutions.identity.users.wsas.WSASRealm;
import org.wso2.usermanager.Realm;
import org.wso2.usermanager.custom.jdbc.JDBCRealm;
import org.wso2.usermanager.custom.jdbc.JDBCRealmConfig;
import org.wso2.usermanager.custom.ldap.LDAPRealm;
import org.wso2.usermanager.custom.ldap.LDAPRealmConfig;
import org.wso2.usermanager.readwrite.DefaultRealm;
import org.wso2.usermanager.readwrite.DefaultRealmConfig;
import org.wso2.usermanager.verification.email.EmailVerifier;
import org.wso2.usermanager.verification.email.EmailVerifierConfig;
import org.wso2.utils.ServerConfiguration;
import org.wso2.utils.ServerException;
import org.wso2.wsas.ServerInitializer;
import org.wso2.wsas.admin.service.UserAdmin;

/**
 * WSO2 WSAS <code>ServerInitializer</code> implementation to carry out
 * initial configuration setup of the Identity Solution : Identity Provider.
 */
public class Initializer implements ServerInitializer {

    private static Log log = LogFactory.getLog(Initializer.class);

    private static Messages messages = Messages
            .getInstance(IdentityProviderConstants.RESOURCES);

    /**
     * {@inheritDoc} Here we carry out all initialization work of the identity
     * solution : identity provider
     */
    public void init(ConfigurationContext ctx) throws AxisFault,
            ServerException {
        log.info(messages.getMessage("initStart"));

        ConfigurationContextHolder.setConfigurationContext(ctx);

        try {
            IPPersistenceManager db = IPPersistenceManager
                    .getPersistanceManager();

            // Checking whether the initial startup - at least one dialect must
            // be there
            DialectDO[] dialectDOs = db.getAllSupportedDialects();
            boolean isInitial = false;
            if (dialectDOs.length == 0) {
                isInitial = true;
            }

            STSConfigAdmin.configuraServices(isInitial);

            if (isInitial) {
                addDialectsAndClaims();

                setupRelyingPartyUserAccounts();

                addParameters();

                setupAllRealms();

                addActions();

                // Add the globally trusted relying party
                RelyingPartyDO rp = new RelyingPartyDO();

                ServerConfiguration serverConfig = ServerConfiguration
                        .getInstance();
                rp.setHostName(serverConfig.getFirstProperty("HostName"));
                rp.setAlias(serverConfig
                        .getFirstProperty("Security.KeyStore.KeyAlias"));
                db.create(rp);

            }

            initializeEmailVerifier();

        } catch (Exception e) {
            throw new AxisFault(e.getMessage(), e);
        }

        log.info(messages.getMessage("initDone"));

    }

    private void setupAllRealms() throws Exception {
        IPPersistenceManager db = IPPersistenceManager.getPersistanceManager();
        // Add JDBC realm
        RealmDO jdbcRealm = new RealmDO();
        jdbcRealm.setClassName(JDBCRealm.class.getName());
        jdbcRealm.setConfigClassName(JDBCRealmConfig.class.getName());
        db.create(jdbcRealm);

        // Add LDAP realm
        RealmDO ldapRealm = new RealmDO();
        ldapRealm.setClassName(LDAPRealm.class.getName());
        ldapRealm.setConfigClassName(LDAPRealmConfig.class.getName());
        db.create(ldapRealm);

        RealmDO wsasRealm = new RealmDO();
        wsasRealm.setClassName(WSASRealm.class.getName());
        wsasRealm.setConfigClassName(null);
        db.create(wsasRealm);

        RealmConfigurationDO wsasRealmConfig = new RealmConfigurationDO();
        wsasRealmConfig.setName("wsasRealmForAccessControlSample");
        wsasRealmConfig.setRealm(wsasRealm);
        db.create(wsasRealmConfig);

        RealmDO defaultRealm = new RealmDO();
        defaultRealm.setClassName(IdentityDefaultRealm.class.getName());
        defaultRealm.setConfigClassName(DefaultRealmConfig.class.getName());
        db.create(defaultRealm);

        // Add default realm configuration
        RealmConfigurationDO realmConfig = new RealmConfigurationDO();
        realmConfig.setName("defaultRealm");
        realmConfig.setRealm(defaultRealm);
        realmConfig.setEffective(true);
        db.create(realmConfig);

        RealmConfigurationPropertyDO prop = new RealmConfigurationPropertyDO();
        prop.setConfig(realmConfig);
        prop.setName("ConnectionURL");
        prop.setValue(IdentityProviderConstants.Default.CONNECTION_URL);
        db.create(prop);

        prop = new RealmConfigurationPropertyDO();
        prop.setConfig(realmConfig);
        prop.setName("ConnectionUserName");
        prop.setValue(IdentityProviderConstants.Default.CONNECTION_USER_NAME);
        db.create(prop);

        prop = new RealmConfigurationPropertyDO();
        prop.setConfig(realmConfig);
        prop.setName("ConnectionPassword");
        prop.setValue(IdentityProviderConstants.Sample.CONNECTION_PASSWORD);
        db.create(prop);

        prop = new RealmConfigurationPropertyDO();
        prop.setConfig(realmConfig);
        prop.setName("DriverName");
        prop.setValue(IdentityProviderConstants.Default.DRIVER);
        db.create(prop);

        this.setupSampleRealmConfig(db, jdbcRealm);

        // Enable user registration
        ParameterAdmin paramAdmin = new ParameterAdmin();
        paramAdmin.createOrUpdatearameter(
                IdentityProviderConstants.PARAM_NAME_ALLOW_USER_REGISTRATION,
                null);

        // Enable OpenID registration
        paramAdmin
                .createOrUpdatearameter(
                        IdentityProviderConstants.PARAM_NAME_ENABLE_OPENID_REGISTRATION,
                        null);

    }

    /**
     * Setup the user store for the identity provider sample.
     * @param db
     * @param jdbcRealm
     * @throws Exception
     */
    private void setupSampleRealmConfig(IPPersistenceManager db,
            RealmDO jdbcRealm) throws Exception {

        // Add sample realm configuration
        RealmConfigurationDO realmConfig = new RealmConfigurationDO();
        realmConfig.setName("sampleRealm");
        realmConfig.setRealm(jdbcRealm);
        db.create(realmConfig);

        RealmConfigurationPropertyDO prop = new RealmConfigurationPropertyDO();
        prop.setConfig(realmConfig);
        prop.setName("ConnectionURL");
        prop.setValue(IdentityProviderConstants.Sample.CONNECTION_URL);
        db.create(prop);

        prop = new RealmConfigurationPropertyDO();
        prop.setConfig(realmConfig);
        prop.setName("ConnectionUserName");
        prop.setValue(IdentityProviderConstants.Sample.CONNECTION_USER_NAME);
        db.create(prop);

        prop = new RealmConfigurationPropertyDO();
        prop.setConfig(realmConfig);
        prop.setName("ConnectionPassword");
        prop.setValue(IdentityProviderConstants.Sample.CONNECTION_PASSWORD);
        db.create(prop);

        prop = new RealmConfigurationPropertyDO();
        prop.setConfig(realmConfig);
        prop.setName("DriverName");
        prop.setValue(IdentityProviderConstants.Sample.DRIVER);
        db.create(prop);

        prop = new RealmConfigurationPropertyDO();
        prop.setConfig(realmConfig);
        prop.setName("UserTable");
        prop.setValue(IdentityProviderConstants.Sample.USER_TABLE);
        db.create(prop);

        prop = new RealmConfigurationPropertyDO();
        prop.setConfig(realmConfig);
        prop.setName("UserNameColumn");
        prop.setValue(IdentityProviderConstants.Sample.USER_NAME_COLUMN);
        db.create(prop);

        prop = new RealmConfigurationPropertyDO();
        prop.setConfig(realmConfig);
        prop.setName("UserCredentialColumn");
        prop.setValue(IdentityProviderConstants.Sample.PASSWORD_COLUMN);
        db.create(prop);

        prop = new RealmConfigurationPropertyDO();
        prop.setConfig(realmConfig);
        prop.setName("ColumnNames");
        prop.setValue(IdentityProviderConstants.Sample.COLUMN_NAME_EMAIL + ","
                + IdentityProviderConstants.Sample.COLUMN_NAME_FIRSTNAME + ","
                + IdentityProviderConstants.Sample.COLUMN_NAME_LASTNAME);
        db.create(prop);

    }

    private void addDialectsAndClaims() throws IdentityProviderException,
            AxisFault {
        IPPersistenceManager db = IPPersistenceManager.getPersistanceManager();
        if (log.isDebugEnabled()) {
            log.debug(messages.getMessage("addingInitialClaimsInfo"));
        }

        InitialClaimsProcessor claimsProcessor = new InitialClaimsProcessor();
        claimsProcessor
                .process(IdentityProviderConstants.INITIAL_CLAIMS_FILE_PATH);
        ArrayList dialects = claimsProcessor.getDialectList();
        for (Iterator iter = dialects.iterator(); iter.hasNext();) {
            DialectDO dialect = (DialectDO) iter.next();
            db.create(dialect);
        }

        ArrayList claims = claimsProcessor.getClaimList();
        for (Iterator iter = claims.iterator(); iter.hasNext();) {
            ClaimDO claim = (ClaimDO) iter.next();
            if (claim.getUri().equals(IdentityConstants.CLAIM_PPID)) {
                claim.setUserEditable(false);
            }
            db.create(claim);
        }

        // Setup claims for the sample
        ClaimsAdmin claimsAdmin = new ClaimsAdmin();
        ClaimDO[] claimDOs = claimsAdmin.getAllClaims();
        // Set supported
        for (int i = 0; i < claimDOs.length; i++) {

            if (log.isDebugEnabled()) {
                log.debug(messages.getMessage("enableClaimsForSample"));
            }

            ClaimDO claim = claimDOs[i];

            // Following claims or a subset of it is required by the OpenID
            // Simple Attribute Registration - So, we accept all these as
            // required parameters, at a time user registers with us - still
            // this configurable through the Admin UI.

            if (claim.getUri().equals(IdentityConstants.CLAIM_EMAIL_ADDRESS)) {
                claim.setAttrId(IdentityConstants.CLAIM_EMAIL_ADDRESS);
                claim.setRequired(true);
            } else if (claim.getUri().equals(IdentityConstants.CLAIM_NICKNAME)) {
                claim.setAttrId(IdentityConstants.CLAIM_NICKNAME);
            } else if (claim.getUri().equals(IdentityConstants.CLAIM_DOB)) {
                claim.setAttrId(IdentityConstants.CLAIM_DOB);
            } else if (claim.getUri().equals(IdentityConstants.CLAIM_GENDER)) {
                claim.setAttrId(IdentityConstants.CLAIM_GENDER);
            } else if (claim.getUri().equals(
                    IdentityConstants.CLAIM_POSTAL_CODE)) {
                claim.setAttrId(IdentityConstants.CLAIM_POSTAL_CODE);
            } else if (claim.getUri().equals(IdentityConstants.CLAIM_COUNTRY)) {
                claim.setAttrId(IdentityConstants.CLAIM_COUNTRY);
                claim.setRequired(true);
            } else if (claim.getUri().equals(IdentityConstants.CLAIM_LANGUAGE)) {
                claim.setAttrId(IdentityConstants.CLAIM_LANGUAGE);
            } else if (claim.getUri().equals(IdentityConstants.CLAIM_TIME_ZONE)) {
                claim.setAttrId(IdentityConstants.CLAIM_TIME_ZONE);
            } else if (claim.getUri().equals(IdentityConstants.CLAIM_FULLNAME)) {
                claim.setAttrId(IdentityConstants.CLAIM_FULLNAME);
            } else if (claim.getUri()
                    .equals(IdentityConstants.CLAIM_GIVEN_NAME)) {
                claim.setAttrId(IdentityConstants.CLAIM_GIVEN_NAME);
                claim.setRequired(true);
            } else if (claim.getUri().equals(IdentityConstants.CLAIM_SURNAME)) {
                claim.setAttrId(IdentityConstants.CLAIM_SURNAME);
                claim.setRequired(true);
            } else if (claim.getUri().equals(IdentityConstants.CLAIM_PPID)) {
                claim.setRequired(true);
            } else if (claim.getUri().equals(IdentityConstants.CLAIM_OPENID)) {
                claim.setAttrId(IdentityConstants.CLAIM_OPENID);
                claim.setUserEditable(false);
                claimsAdmin.updateClaim(claim);
            }

            if (claim.getUri().equals(IdentityConstants.CLAIM_EMAIL_ADDRESS)
                    || claim.getUri().equals(IdentityConstants.CLAIM_NICKNAME)
                    || claim.getUri().equals(IdentityConstants.CLAIM_DOB)
                    || claim.getUri().equals(IdentityConstants.CLAIM_GENDER)
                    || claim.getUri().equals(
                            IdentityConstants.CLAIM_POSTAL_CODE)
                    || claim.getUri().equals(IdentityConstants.CLAIM_COUNTRY)
                    || claim.getUri().equals(IdentityConstants.CLAIM_LANGUAGE)
                    || claim.getUri().equals(IdentityConstants.CLAIM_TIME_ZONE)
                    || claim.getUri().equals(IdentityConstants.CLAIM_FULLNAME)
                    || claim.getUri()
                            .equals(IdentityConstants.CLAIM_GIVEN_NAME)
                    || claim.getUri().equals(IdentityConstants.CLAIM_SURNAME)
                    || claim.getUri().equals(IdentityConstants.CLAIM_PPID)) {
                claim.setSupported(true);
                claimsAdmin.updateClaim(claim);
            }
        }

    }

    /**
     * Called only when the Identity Solution is coming up for the first time
     */
    private void setupRelyingPartyUserAccounts() throws Exception {
        UserAdmin admin = new UserAdmin();
        String[] roleNames = admin.getRoleNames();
        boolean found = false;
        for (int i = 0; i < roleNames.length; i++) {
            if (IdentityProviderConstants.RP_USER_ROLE.equals(roleNames[i])) {
                found = true;
                break;
            }
        }
        if (!found) {
            admin.addRole(IdentityProviderConstants.RP_USER_ROLE,
                    "Relying Party Users");
        }

    }

    private void initializeEmailVerifier() throws Exception {
        ParameterAdmin admin = new ParameterAdmin();
        boolean enableEmailVerification = admin
                .getParameter(IdentityProviderConstants.PARAM_NAME_ENABLE_EMAIL_VERIFICATION) != null;

        if (!enableEmailVerification) {
            return;
        }
        String value = admin
                .getParameterValue(IdentityProviderConstants.PARAM_NAME_EMAIL_PORT);
        int port = 0;

        if (value != null) {
            port = Integer.parseInt(value);
        }

        String host = admin
                .getParameterValue(IdentityProviderConstants.PARAM_NAME_EMAIL_HOST);
        String fromAddress = admin
                .getParameterValue(IdentityProviderConstants.PARAM_NAME_EMAIL_FROM_ADDRESS);
        String username = admin
                .getParameterValue(IdentityProviderConstants.PARAM_NAME_EMAIL_USERNAME);
        String password = admin
                .getParameterValue(IdentityProviderConstants.PARAM_NAME_EMAIL_PASSWORD);

        UserStore store = UserStore.getInstance();
        Realm realm = store.getRealm();

        EmailVerifierConfig config = new EmailVerifierConfig();
        config.setFromAddress(fromAddress);
        config.setHost(host);
        if (username != null && username.length() != 0) {
            config.setPassword(password);
            config.setUsername(username);
        }

        String verificationAddress = admin
                .getParameterValue(IdentityProviderConstants.PARAM_NAME_EMAIL_VERIFICATION_EPR);

        config.setRegistrationServiceEPR(verificationAddress);
        config.setPort(port);
        if (realm.getClass().getName().equals(DefaultRealm.class.getName())) {
            EmailVerifier.init(realm, config);
        } else {
            throw new IdentityProviderException("emailNotActivated");
        }

    }

    /**
     * Called only when the Identity Solution is coming up for the first time
     */
    private void addParameters() throws Exception {
        ParameterAdmin admin = new ParameterAdmin();

        admin.createOrUpdatearameter(
                IdentityConstants.PARAM_SUPPORTED_TOKEN_TYPES,
                IdentityConstants.SAML10_URL + ","
                        + IdentityConstants.SAML11_URL + ","
                        + IdentityConstants.SAML20_URL + ","
                        + IdentityConstants.OpenId.OPENID_URL);

        admin.createOrUpdatearameter(IdentityConstants.PARAM_CARD_NAME,
                IdentityConstants.PARAM_VALUE_CARD_NAME);
        admin.createOrUpdatearameter(IdentityConstants.PARAM_VALID_PERIOD,
                IdentityConstants.PARAM_VALUE_VALID_PERIOD);

    }

    private void addActions() throws Exception {
        IPPersistenceManager db = IPPersistenceManager.getPersistanceManager();
        for (int i = 0; i < ActionDO.actions.length; i++) {
            ActionDO act = new ActionDO();
            act.setActionId(ActionDO.actions[i]);
            db.create(act);
        }
    }

}
