/*                                                                             
 * Copyright 2005,2006 WSO2, Inc. http://www.wso2.org
 *                                                                             
 * Licensed under the Apache License, Version 2.0 (the "License");             
 * you may not use this file except in compliance with the License.            
 * You may obtain a copy of the License at                                     
 *                                                                             
 *      http://www.apache.org/licenses/LICENSE-2.0                             
 *                                                                             
 * Unless required by applicable law or agreed to in writing, software         
 * distributed under the License is distributed on an "AS IS" BASIS,           
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.    
 * See the License for the specific language governing permissions and         
 * limitations under the License.                                              
 */

package org.wso2.solutions.identity.sts;

import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMFactory;
import org.apache.axiom.om.OMNode;
import org.apache.axiom.om.util.UUIDGenerator;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axis2.context.MessageContext;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.rahas.RahasConstants;
import org.apache.rahas.RahasData;
import org.apache.rahas.Token;
import org.apache.rahas.TokenIssuer;
import org.apache.rahas.TrustException;
import org.apache.rahas.TrustUtil;
import org.apache.ws.security.WSConstants;
import org.apache.ws.security.message.WSSecEncryptedKey;
import org.apache.ws.security.message.token.SecurityTokenReference;
import org.apache.ws.security.util.WSSecurityUtil;
import org.apache.ws.security.util.XmlSchemaDateFormat;
import org.apache.xml.security.encryption.EncryptedData;
import org.apache.xml.security.encryption.XMLCipher;
import org.apache.xml.security.keys.KeyInfo;
import org.joda.time.DateTime;
import org.opensaml.SAMLException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.wso2.solutions.identity.IdentityConstants;
import org.wso2.solutions.identity.IdentityProviderConstants;
import org.wso2.solutions.identity.IdentityProviderException;
import org.wso2.solutions.identity.admin.ParameterAdmin;
import org.wso2.solutions.identity.admin.ReportAdmin;
import org.wso2.solutions.identity.i18n.Messages;
import org.wso2.solutions.identity.persistence.IPPersistenceManager;
import org.wso2.solutions.identity.persistence.dataobject.ActionDO;
import org.wso2.solutions.identity.persistence.dataobject.InfoCardDO;
import org.wso2.solutions.identity.persistence.dataobject.IssuedTokensDO;
import org.wso2.solutions.identity.sts.IdentityProviderData.RequestedClaimData;
import org.wso2.solutions.identity.sts.saml.SAML1TokenBuilder;
import org.wso2.solutions.identity.sts.saml.SAML2TokenBuilder;
import org.wso2.solutions.identity.sts.saml.SAMLTokenBuilder;
import org.wso2.solutions.identity.sts.saml.SAMLTokenDirector;
import org.wso2.solutions.identity.util.IdentityUtil;

import javax.crypto.SecretKey;
import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilderFactory;

import java.io.ByteArrayInputStream;
import java.security.cert.X509Certificate;
import java.text.DateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Iterator;
import java.util.Map;
import java.util.TimeZone;

/**
 * This is the main token issuer implementation identity provider. Two types of
 * credential mechanisms are supported:
 * 
 * <ul>
 * <li>Username and Password Credential</li>
 * <li>Self-issued Token Credential</li>
 * </ul>
 * 
 * Supported list of claims can be configured and the values of those supported
 * claims will be obtained from the user store setup to work with the identity
 * provider.
 */
public class IdentityTokenIssuer implements TokenIssuer {

    private static Log log = LogFactory.getLog(IdentityTokenIssuer.class);

    private final static String WSS_SAML_NS = "http://docs.oasis-open.org/wss/oasis-wss-saml-token-profile-1.0#";

    protected IdentityProviderData ipData = null;

    public final static String ISSUER_SELF = IdentityConstants.NS
            + "/issuer/self";

    private static Log tokenIssuerLog = LogFactory
            .getLog(IdentityProviderConstants.TOKEN_ISSUSER_LOG);

    private static Messages messages = Messages
            .getInstance(IdentityProviderConstants.RESOURCES);

    private boolean isTokenLogDebug = false;

    public IdentityTokenIssuer() {
        isTokenLogDebug = tokenIssuerLog.isDebugEnabled();
    }

    /**
     * {@inheritDoc}
     */
    public String getResponseAction(RahasData data) throws TrustException {
        return RahasConstants.WST_NS_05_02 + RahasConstants.RSTR_ACTION_ISSUE;
    }

    /**
     * {@inheritDoc}
     */
    public SOAPEnvelope issue(RahasData data) throws TrustException {

        boolean debug = log.isDebugEnabled();

        if (debug) {
            log.debug("issue");
        }
        try {
            if (debug) {
                log.debug("Request: \n" + data.getRstElement().toString()
                        + "\n\n");
            }

            ipData = getIdentityProviderData(data);
            boolean isValidCard = isValidCard(ipData.getCardID());

            if (debug) {
                log.debug("Card is validated");
            }

            if (isValidCard == false) {
                String msg = "Invalid information card";
                log.error(msg);
                ReportAdmin.record(ipData.getUserIdentifier(),
                        ActionDO.ACTION_TOKEN_ISSUE_FAILURE, msg);
                throw new TrustException(TrustException.REQUEST_FAILED);

            }

            if (isTokenLogDebug) {
                tokenIssuerLog.debug(messages.getMessage("validInfoCard",
                        new String[] { ipData.getCardID() }));
            }

            return createResponse(data);

        } catch (Exception e) {
            throw new TrustException(TrustException.REQUEST_FAILED, e);
        } finally {
            log.info("Issued token");
        }

    }

    /**
     * Create the response SOAP envelope.
     * 
     * @param data
     *            WS-Trust information in the issue request.
     * @return response SOAP envelope.
     * @throws TrustException
     */
    private SOAPEnvelope createResponse(RahasData rahasData)
            throws TrustException {

        try {

            MessageContext inMsgCtx = rahasData.getInMessageContext();

            SOAPEnvelope env = TrustUtil.createSOAPEnvelope(inMsgCtx
                    .getEnvelope().getNamespace().getNamespaceURI());
            Document doc = ((Element) env).getOwnerDocument();

            // Create EncryptedKey
            WSSecEncryptedKey encryptedKey = null;

            X509Certificate serviceCert = ipData.getRpCert();
            if (serviceCert != null) {
                encryptedKey = new WSSecEncryptedKey();
                encryptedKey.setUseThisCert(serviceCert);
                encryptedKey.setKeySize(256);
                encryptedKey.setKeyEncAlgo(WSConstants.KEYTRANSPORT_RSAOEP);
                encryptedKey
                        .setKeyIdentifierType(WSConstants.THUMBPRINT_IDENTIFIER);
                encryptedKey.prepare(doc, null);

                Element encrKeyElem = encryptedKey.getEncryptedKeyElement();

                // Create KeyInfo structure - START

                Element keyInfoElem = doc.createElementNS(WSConstants.SIG_NS,
                        "KeyInfo");
                ((OMElement) encrKeyElem).declareNamespace(WSConstants.SIG_NS,
                        WSConstants.SIG_PREFIX);
                ((OMElement) encrKeyElem).declareNamespace(WSConstants.ENC_NS,
                        WSConstants.ENC_PREFIX);
                keyInfoElem.appendChild(encrKeyElem);

                // Create KeyInfo structure - END

            }

            if (!checkIsValidTokenType(ipData)) {
                throw new IdentityProviderException("invalidTokenType",
                        new Object[] { ipData.getRequiredTokenType() });
            }

            DateTime notBefore = new DateTime();
            DateTime notAfter = new DateTime(notBefore.getMillis()
                    + (300 * 1000));
            String assertionId = UUIDGenerator.getUUID();

            if (isTokenLogDebug) {
                tokenIssuerLog.debug(messages.getMessage(
                        "startSAMLTokenCreation", new String[] { ipData
                                .getCardID() }));
            }
            Element assertionNode = createSAMLAssertionAsDOM(ipData, rahasData,
                    notBefore, notAfter, assertionId);

            if (isTokenLogDebug) {
                tokenIssuerLog.debug(messages.getMessage(
                        "finishSAMLTokenCreation", new String[] { ipData
                                .getCardID() }));
            }

            OMElement rstrElem = createRSTR(rahasData, notBefore.toDate(),
                    notAfter.toDate(), env, doc, assertionNode, assertionId,
                    encryptedKey);

            if (isTokenLogDebug) {
                tokenIssuerLog.debug(messages.getMessage("RSTRCreationDone",
                        new String[] { ipData.getCardID() }));
            }
            if (log.isDebugEnabled()) {
                log.debug("Response created");
                log.debug("Response body : \n" + rstrElem.toString() + "\n\n");
            }
            IPPersistenceManager dbman = IPPersistenceManager
                    .getPersistanceManager();
            InfoCardDO card = dbman.getInfoCard(ipData.getCardID());

            IssuedTokensDO tok = new IssuedTokensDO();
            tok.setCard(card);
            tok.setDateExpires(notAfter.toDate());
            tok.setDateIssued(notBefore.toDate());
            
            if (rahasData.getTokenType() == null
                    || rahasData.getTokenType().trim().length() == 0) {
                tok.setTokenType(ipData.getDefautTokenType());
            } else {
                tok.setTokenType(rahasData.getTokenType());
            }

            dbman.create(tok);
            return env;

        } catch (Exception e) {
            log.error(e.getMessage());
            try {
                ReportAdmin.record(ipData.getUserIdentifier(),
                        ActionDO.ACTION_TOKEN_ISSUE_FAILURE, e.getMessage());
            } catch (IdentityProviderException e1) {
                throw new TrustException(TrustException.REQUEST_FAILED, e1);
            }
            throw new TrustException(TrustException.REQUEST_FAILED, e);
        } finally {
            log.info("Response ready for : " + ipData.getCardID());
        }
    }

    /**
     * Create the <code>wst:RequstedSecurityTokenRespoonse</code> element.
     * 
     * @param data
     *            WS-Trust information in the issue request
     * @param notBefore
     *            Created time
     * @param notAfter
     *            Expiration time
     * @param env
     *            Response SOAP envelope
     * @param doc
     *            <code>org.w3.dom.Document</code> instance of the response
     *            SOAP envelope
     * @param assertion
     *            SAML Assertion to be sent in the response.
     * @param encryptedKey
     *            Key used to encrypt the SAML assertion.
     * @return <code>wst:RequstedSecurityTokenRespoonse</code> element.
     * @throws TrustException
     * @throws SAMLException
     */
    protected OMElement createRSTR(RahasData data, Date notBefore,
            Date notAfter, SOAPEnvelope env, Document doc, Node assertionElem,
            String assertionId, WSSecEncryptedKey encryptedKey)
            throws TrustException, SAMLException, IdentityProviderException {
        if (log.isDebugEnabled()) {
            log.debug("Begin RSTR Element creation.");
        }
        int wstVersion = data.getVersion();
        MessageContext inMsgCtx = data.getInMessageContext();

        OMElement rstrElem = TrustUtil
                .createRequestSecurityTokenResponseElement(wstVersion, env
                        .getBody());
        TrustUtil.createTokenTypeElement(wstVersion, rstrElem).setText(
                data.getTokenType());

        createDisplayToken(rstrElem, ipData);

        OMElement appliesToEpr = null;
        if (encryptedKey != null) {
            int keysize = data.getKeysize();
            if (keysize == -1) {
                keysize = encryptedKey.getEphemeralKey().length * 8;
            }

            TrustUtil.createKeySizeElement(wstVersion, rstrElem, keysize);

            OMElement incomingAppliesToEpr = data.getAppliesToEpr();
            try {
                Document eprDoc = DocumentBuilderFactory.newInstance()
                        .newDocumentBuilder().parse(
                                new ByteArrayInputStream(incomingAppliesToEpr
                                        .toString().getBytes()));
                appliesToEpr = (OMElement) doc.importNode(eprDoc
                        .getDocumentElement(), true);
            } catch (Exception e) {
                throw new TrustException(TrustException.REQUEST_FAILED, e);
            }

            OMElement appliesToElem = rstrElem
                    .getOMFactory()
                    .createOMElement(
                            new QName(
                                    RahasConstants.WSP_NS,
                                    RahasConstants.IssuanceBindingLocalNames.APPLIES_TO,
                                    RahasConstants.WSP_PREFIX), rstrElem);
            appliesToElem.addChild(appliesToEpr);
        }

        // Use GMT time in milliseconds
        DateFormat zulu = new XmlSchemaDateFormat();

        // Add the Lifetime element
        TrustUtil.createLifetimeElement(wstVersion, rstrElem, zulu
                .format(notBefore), zulu.format(notAfter));

        OMElement reqSecTokenElem = TrustUtil
                .createRequestedSecurityTokenElement(wstVersion, rstrElem);

        Node assertionElement = doc.importNode(assertionElem, true);
        reqSecTokenElem.addChild((OMNode) assertionElement);

        if (log.isDebugEnabled()) {
            log.debug(assertionElement.toString());
        }

        if (encryptedKey != null) {
            encryptSAMLAssertion(doc, (Element) assertionElement, encryptedKey);
        }

        createAttachedRef(rstrElem, assertionId);
        createUnattachedRef(rstrElem, assertionId);

        // Store the Token
        Token assertionToken = new Token(assertionId, (OMElement) doc
                .importNode(assertionElem, true), notBefore, notAfter);

        // At this point we definitely have the secret
        // Otherwise it should fail with an exception earlier
        assertionToken.setSecret(data.getEphmeralKey());
        TrustUtil.getTokenStore(inMsgCtx).add(assertionToken);

        // Creating the ReqProoftoken - END
        if (log.isDebugEnabled()) {
            log.debug("RSTR Elem created.");
        }
        log.info("RSTR ready with token : " + assertionId);
        return rstrElem;
    }

    /**
     * Create and add wst:AttachedReference element
     * 
     * @param rstrElem
     *            wst:RequestSecurityToken element
     * @param id
     *            Token identifier
     */
    protected void createAttachedRef(OMElement rstrElem, String id) {
        OMFactory fac = rstrElem.getOMFactory();
        OMElement rar = fac
                .createOMElement(
                        new QName(
                                RahasConstants.WST_NS_05_02,
                                RahasConstants.IssuanceBindingLocalNames.REQUESTED_ATTACHED_REFERENCE,
                                RahasConstants.WST_PREFIX), rstrElem);
        OMElement str = fac.createOMElement(new QName(WSConstants.WSSE_NS,
                SecurityTokenReference.SECURITY_TOKEN_REFERENCE,
                WSConstants.WSSE_PREFIX), rar);
        OMElement ki = fac.createOMElement(new QName(WSConstants.WSSE_NS,
                "KeyIdentifier", WSConstants.WSSE_PREFIX), str);
        ki.addAttribute("ValueType", WSS_SAML_NS
                + WSConstants.SAML_ASSERTION_ID, null);
        ki.setText(id);
    }

    /**
     * Create and add wst:UnattachedReference element
     * 
     * @param rstrElem
     *            wst:RequestSecurityToken element
     * @param id
     *            Token identifier
     */
    protected void createUnattachedRef(OMElement rstrElem, String id) {
        OMFactory fac = rstrElem.getOMFactory();
        OMElement rar = fac
                .createOMElement(
                        new QName(
                                RahasConstants.WST_NS_05_02,
                                RahasConstants.IssuanceBindingLocalNames.REQUESTED_UNATTACHED_REFERENCE,
                                RahasConstants.WST_PREFIX), rstrElem);
        OMElement str = fac.createOMElement(new QName(WSConstants.WSSE_NS,
                SecurityTokenReference.SECURITY_TOKEN_REFERENCE,
                WSConstants.WSSE_PREFIX), rar);
        OMElement ki = fac.createOMElement(new QName(WSConstants.WSSE_NS,
                "KeyIdentifier", WSConstants.WSSE_PREFIX), str);

        ki.addAttribute("ValueType", WSS_SAML_NS
                + WSConstants.SAML_ASSERTION_ID, null);
        ki.setText(id);
    }

    /**
     * Encrypt the given SAML Assertion element with the given key information.
     * 
     * @param doc
     * @param assertionElement
     * @param encryptedKey
     */
    private void encryptSAMLAssertion(Document doc, Element assertionElement,
            WSSecEncryptedKey encryptedKey) throws TrustException {
        try {
            XMLCipher xmlCipher = XMLCipher.getInstance(WSConstants.AES_256);
            SecretKey secretKey = WSSecurityUtil.prepareSecretKey(
                    WSConstants.AES_256, encryptedKey.getEphemeralKey());
            xmlCipher.init(XMLCipher.ENCRYPT_MODE, secretKey);
            String xencEncryptedDataId = "EncDataId-"
                    + assertionElement.hashCode();

            KeyInfo keyInfo = new KeyInfo(doc);
            keyInfo.addUnknownElement(encryptedKey.getEncryptedKeyElement());

            EncryptedData encData = xmlCipher.getEncryptedData();
            encData.setId(xencEncryptedDataId);
            encData.setKeyInfo(keyInfo);
            xmlCipher.doFinal(doc, assertionElement, false);
        } catch (Exception e) {
            throw new TrustException(TrustException.REQUEST_FAILED, e);
        }
    }

    /**
     * Create the DisplayToken element according to CardSpace specifications.
     * 
     * @param rahasData
     *            Information from the WS-Trust request.
     * @param ipData
     *            CardSpace specific meta-data for this issuance.
     * @return The DisplayToken element.
     */
    protected OMElement createDisplayToken(OMElement rstrElem,
            IdentityProviderData ipData) throws IdentityProviderException {
        if (log.isDebugEnabled()) {
            log.debug("Begin Display token creation.");
        }

        Map requestedClaims = ipData.getRequestedClaims();

        if (requestedClaims.isEmpty())
            return null;

        OMElement rdt = IdentityProviderUtil.createRequestedDisplayToken(
                rstrElem, ipData);
        OMElement displayToken = IdentityProviderUtil.createDisplayToken(rdt,
                ipData);

        try {
            Iterator ite = requestedClaims.values().iterator();
            while (ite.hasNext()) {
                RequestedClaimData claim = (RequestedClaimData) ite.next();

                if (claim.uri.equals(IdentityConstants.CLAIM_PPID)) {
                    // PPID display token
                    IdentityProviderUtil.createDisplayClaim(displayToken,
                            ipData.getDisplayName(claim.uri), IdentityUtil
                                    .getPPIDDisplayValue(claim.value),
                            claim.uri);
                } else {
                    IdentityProviderUtil.createDisplayClaim(displayToken,
                            ipData.getDisplayName(claim.uri), claim.value,
                            claim.uri);
                }
            }
        } catch (Exception e) {
            throw new IdentityProviderException(e.getMessage(), e);
        }

        if (log.isDebugEnabled()) {
            log.debug("createDisplayToken");
        }

        return rdt;
    }

    /**
     * {@inheritDoc}
     */
    public void setConfigurationElement(OMElement configElement) {
        // Nothing to do
    }

    /**
     * {@inheritDoc}
     */
    public void setConfigurationFile(String configFile) {
        // Nothing to do
    }

    /**
     * {@inheritDoc}
     */
    public void setConfigurationParamName(String configParamName) {
        // Nothing to do
    }

    /**
     * Check whether the information card referenced in the token request is a
     * valid managed information card issued by the identity provider.
     * 
     * @param cardId
     *            Identifier of the information card.
     * @return Whether the card is valid or not.
     * @throws IdentityProviderException
     */
    private boolean isValidCard(String cardId) throws IdentityProviderException {
        if (log.isDebugEnabled()) {
            log.debug("Begin Validating card.");
        }
        boolean retval = false;
        IPPersistenceManager dbman = IPPersistenceManager
                .getPersistanceManager();
        InfoCardDO card = dbman.getInfoCard(cardId);
        if (card != null) {
            Calendar cal = new GregorianCalendar(TimeZone.getTimeZone("UTC"));
            Date now = cal.getTime();

            if (now.before(card.getDateExpires())
                    && now.after(card.getDateIssued())) {
                retval = true;
            }
        }

        return retval;
    }

    protected IdentityProviderData getIdentityProviderData(RahasData rahasData)
            throws Exception {
        return new IdentityProviderData(rahasData);
    }

    protected Element createSAMLAssertionAsDOM(IdentityProviderData ipData,
            RahasData rahasData, DateTime notBefore, DateTime notAfter,
            String assertionId) throws IdentityProviderException {

        Element elem = null;
        SAMLTokenBuilder builder = null;
        final String requiredTokenType = ipData.getRequiredTokenType();
        if (requiredTokenType.equals(IdentityConstants.SAML10_URL)
                || requiredTokenType.equals(IdentityConstants.SAML11_URL)) {
            builder = new SAML1TokenBuilder();
        } else if (requiredTokenType.equals(IdentityConstants.SAML20_URL)) {
            builder = new SAML2TokenBuilder();
        }

        SAMLTokenDirector director = new SAMLTokenDirector(builder, rahasData,
                ipData);
        elem = director.createSAMLToken(notBefore, notAfter, assertionId);

        return elem;
    }

    protected boolean checkIsValidTokenType(IdentityProviderData data)
            throws IdentityProviderException {
        boolean isValid = false;
        String type = data.getRequiredTokenType();
        ParameterAdmin admin = new ParameterAdmin();
        String types = admin
                .getParameterValue(IdentityConstants.PARAM_SUPPORTED_TOKEN_TYPES);
        String[] arrTypes = types.split(",");

        for (int i = 0; i < arrTypes.length; i++) {
            if (arrTypes[i].equals(type)) {
                isValid = true;
                break;
            }
        }
        return isValid;
    }

}
